import React from "react"
import dynamic from "next/dynamic"
import { useRouter } from "next/router"
import { Badge, useColorMode, Wrap, WrapItem } from "@chakra-ui/react"
import EmptyView from "components/utils/EmptyView"
import { TableColumn } from "react-data-table-component"
import { RISK_TO_COLOR } from "~/constants"
import { getCustomStyles, rowStyles } from "components/utils/TableUtils"
import { VulnerabilityAggItem, GetVulnerabilityAggParams } from "@common/types"
const DataTable = dynamic(() => import("react-data-table-component"), {
  ssr: false,
})

const PAGE_SIZE = 10

interface VulnerabilityTableProps {
  items: VulnerabilityAggItem[]
  params: GetVulnerabilityAggParams
}

const List: React.FC<VulnerabilityTableProps> = React.memo(
  ({ items, params }) => {
    const colorMode = useColorMode()
    const router = useRouter()
    const columns: TableColumn<VulnerabilityAggItem>[] = [
      {
        name: "Type",
        sortable: true,
        selector: (row: VulnerabilityAggItem) => row.type,
        id: "type",
      },
      {
        name: "Risk Score",
        sortable: true,
        selector: (row: VulnerabilityAggItem) => row.risk,
        cell: (row: VulnerabilityAggItem) => (
          <Badge
            p="1"
            fontSize="sm"
            colorScheme={RISK_TO_COLOR[row.risk]}
            pointerEvents="none"
          >
            {row.risk}
          </Badge>
        ),
        id: "risk",
      },
      {
        name: "Count",
        sortable: true,
        selector: (row: VulnerabilityAggItem) => row.count,
        cell: (row: VulnerabilityAggItem) => (
          <Wrap
            display="flex"
            alignItems="center"
            h="full"
            pr="5"
            className="my-box"
            cursor="pointer"
          >
            <WrapItem>{row.count}</WrapItem>
            <WrapItem
              onClick={() =>
                router.push({
                  pathname: "/alerts",
                  query: {
                    alertTypes: row.type,
                    hosts: params.hosts?.length
                      ? params.hosts.join(",")
                      : undefined,
                  },
                })
              }
            >
              View All →
            </WrapItem>
          </Wrap>
        ),
        id: "count",
      },
      {
        name: "Endpoints",
        sortable: true,
        selector: (row: VulnerabilityAggItem) => row.numEndpoints,
        id: "numEndpoints",
      },
      {
        name: "Hosts",
        sortable: true,
        selector: (row: VulnerabilityAggItem) => row.numHosts,
        id: "numHosts",
      },
    ]

    if (items.length == 0) {
      return <EmptyView text="No results found." />
    }
    if (items.length > 0) {
      return (
        <DataTable
          style={rowStyles}
          paginationComponentOptions={{ noRowsPerPage: true }}
          paginationTotalRows={items.length}
          paginationPerPage={PAGE_SIZE}
          columns={columns}
          data={items}
          customStyles={getCustomStyles(colorMode.colorMode, false, true)}
          pagination
        />
      )
    }
    return null
  },
)

export default List
